/*
 * Created on Nov 7, 2006
 */
package net.sf.doolin.bus;

import java.util.Collection;

import net.sf.doolin.bus.support.SubscriberValidator;

/**
 * Definition for an event bus.
 * 
 * @author Damien Coraboeuf
 */
public abstract class Bus {

	/**
	 * Default bus
	 */
	private static Bus defaultBus = new DefaultBus();

	/**
	 * Get the default bus
	 * 
	 * @return Default bus
	 */
	public static Bus get() {
		return defaultBus;
	}

	/**
	 * Sets a new default implementation
	 * 
	 * @param bus
	 *            New bus implementation
	 */
	public static void setDefault(Bus bus) {
		defaultBus = bus;
	}

	/**
	 * Cleans the subscribers by checking their validity.
	 */
	public abstract void cleanUp();

	/**
	 * Deletes all subscriptions from this bus.
	 */
	public abstract void clear();

	/**
	 * Gets the bus log
	 */
	public abstract BusLog getBusLog();

	/**
	 * List of <i>all</i> subscribers and their associated meta information.
	 * 
	 * @return All subscribers
	 */
	public abstract Collection<SubscriberInfo> getSubscribers();

	/**
	 * List of subscribers for a message.
	 * 
	 * @param <M>
	 *            Message class.
	 * @param message
	 *            Message.
	 * @return List of subscribers, never <code>null</code> but can be empty.
	 */
	public abstract <M> Collection<Subscriber<M>> getSubscribers(M message);

	/**
	 * Publishes a message
	 * 
	 * @param <Message>
	 *            Message type
	 * @param message
	 *            Message.
	 */
	public abstract <Message> void publish(Message message);

	/**
	 * Sets the log to be used by this bus.
	 * 
	 * @param log
	 *            Bus log instance
	 */
	public abstract void setBusLog(BusLog log);

	/**
	 * Subscribes to a message class.
	 * 
	 * @param <M>
	 *            Message type
	 * @param subscriber
	 *            Subscriber
	 */
	public abstract <M> void subscribe(Subscriber<M> subscriber);

	/**
	 * Subscribes to a message class.
	 * 
	 * @param <M>
	 *            Message type
	 * @param subscriber
	 *            Subscriber
	 * @param priority
	 *            Level of priority (high priorities are handled first)
	 */
	public abstract <M> void subscribe(Subscriber<M> subscriber,
			Priority priority);

	/**
	 * Registers the components of a subscription to this bus.
	 * 
	 * @param <M>
	 *            Type of message
	 * @param validator
	 *            Validator for this subscription
	 * @param trigger
	 *            Trigger for this subscription
	 * @param execution
	 *            Execution for this subscription
	 */
	public abstract <M> void subscribe(SubscriberValidator validator,
			SubscriberTrigger trigger, SubscriberExecution<M> execution);

	/**
	 * Unregister a subscriber.
	 * 
	 * @param subscriber
	 *            Subscriber to unregister.
	 */
	public abstract void unsubscribe(Subscriber<?> subscriber);

}
