package net.sf.doolin.bus;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Message processor registration
 * 
 * @param <M>
 *            Message type
 */
public class Subscription<M> implements Comparable<Subscription<M>> {
	/**
	 * Logger
	 */
	private final static Logger log = LoggerFactory
			.getLogger(Subscription.class);

	/**
	 * Associated message processor
	 */
	private final Subscriber<M> subscriber;

	/**
	 * Priority
	 */
	private final Priority priority;

	/**
	 * Registration time
	 */
	private final long registrationTime;

	/**
	 * Constructor
	 * 
	 * @param subscriber
	 *            Associated subscriber
	 */
	public Subscription(Subscriber<M> subscriber) {
		this(subscriber, Priority.NORMAL);
	}

	/**
	 * Constructor
	 * 
	 * @param subscriber
	 *            Associated subscriber
	 * @param priority
	 *            Priority for the message
	 */
	public Subscription(Subscriber<M> subscriber, Priority priority) {
		this.subscriber = subscriber;
		this.priority = priority;
		this.registrationTime = System.currentTimeMillis();
	}

	/**
	 * Checks if this subscription accepts a message
	 * 
	 * @param message
	 *            Message to test
	 * @return Result of the test
	 */
	public boolean accept(Object message) {
		return this.subscriber.accept(message);
	}

	/**
	 * Comparison is based on the priority. But two priorities but not been
	 * equals.
	 */
	public int compareTo(Subscription<M> o) {
		if (o.priority == this.priority) {
			return 1;
		} else {
			/*
			 * THIS has more priority than O if its priority is hight
			 */
			return (o.priority.compareTo(this.priority));
		}
	}

	/**
	 * @see java.lang.Object#equals(Object)
	 */
	@Override
	public boolean equals(Object o) {
		Subscription<?> registration = (Subscription<?>) o;
		boolean ok = this.subscriber == registration.subscriber;
		if (log.isTraceEnabled()) {
			log.trace("equals " + ok + " for " + o + " and " + this);
		}
		return ok;
	}

	/**
	 * Returns the priority.
	 * 
	 * @return int
	 */
	public Priority getPriority() {
		return this.priority;
	}

	/**
	 * Returns the registration date
	 * 
	 * @return Registration date
	 * @see System#currentTimeMillis()
	 */
	public long getRegistrationTime() {
		return this.registrationTime;
	}

	/**
	 * @return Subscriber
	 */
	public Subscriber<M> getSubscriber() {
		return this.subscriber;
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return this.subscriber.hashCode();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return this.subscriber.toString();
	}
}
