package net.sf.doolin.bus.bean;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;

import net.sf.doolin.bus.support.PropertyChangeMessage;

/**
 * This class defines an object that can notify its changes to the bus.
 * <p>
 * Notification of changes can be done on the spot, or buffered or completely
 * stopped.
 * <p>
 * Any {@link Bean} can have a parent and we can therefore define a hierarchy of
 * {@linkplain Bean beans}. Each level in the hierarchy participates in the
 * notification of changes.
 * <p>
 * For example, in the following hierarchy:
 * 
 * <pre>
 * A {
 * 	B b;
 * }
 * 
 * B {
 * 	C c;
 * }
 * 
 * C {
 * 	String name;
 * }
 * </pre>
 * 
 * If C notifies a change on <code>name</code> as in:
 * 
 * <pre>
 * C {
 * 	...
 * 	notify(&quot;name&quot;, &quot;Old value&quot;, &quot;A new name&quot;);
 * 	...
 * }
 * </pre>
 * 
 * Following events are generated on the bus:
 * 
 * <pre>
 * C.name := &quot;A new name&quot;;
 * B.c.name := &quot;A new name&quot;;
 * A.b.c.name := &quot;A new name&quot;;
 * </pre>
 * 
 * @author Damien Coraboeuf
 * 
 */
public abstract class Bean implements Serializable, IBean {

	private static final long serialVersionUID = 1L;

	private transient IParentBean parentBean;
	private transient BeanNotificationSupport notificationSupport;

	public Bean() {
		this.notificationSupport = new BeanNotificationSupport(this);
	}

	/**
	 * @see BeanNotificationSupport#bufferedNotification()
	 */
	public NotificationMode bufferedNotification() {
		return this.notificationSupport.bufferedNotification();
	}

	/**
	 * If the bean is in buffer mode, cancels all changes.
	 * 
	 * @see BeanNotificationSupport#cancelBuffer()
	 */
	public void cancelBuffer() {
		this.notificationSupport.cancelBuffer();
	}

	/**
	 * Cancels all buffered changes and restore the buffer state
	 * 
	 * @see BeanNotificationSupport#cancelBufferAndRestoreBuffer()
	 */
	public void cancelBufferAndRestoreBuffer() {
		this.notificationSupport.cancelBufferAndRestoreBuffer();
	}

	/**
	 * @see BeanNotificationSupport#defaultNotification()
	 */
	public NotificationMode defaultNotification() {
		return this.notificationSupport.defaultNotification();
	}

	@Override
	public BeanNotificationSupport getBeanNotificationSupport() {
		return this.notificationSupport;
	}

	/**
	 * @see BeanNotificationSupport#getNotificationMode()
	 */
	public NotificationMode getNotificationMode() {
		return this.notificationSupport.getNotificationMode();
	}

	@Override
	public IParentBean getParentBean() {
		return this.parentBean;
	}

	/**
	 * @see BeanNotificationSupport#isHasBufferedChanges()
	 */
	public boolean isHasBufferedChanges() {
		return this.notificationSupport.isHasBufferedChanges();
	}

	/**
	 * Notifies to the bus that a simple property (non-{@link Bean}) has
	 * changed. If the notification mode is BUFFERED, the change will be
	 * actually sent when notification mode is restored to NORMAL.
	 * 
	 * @param propertyName
	 *            Name of the property
	 * @param oldValue
	 *            Old value of the property
	 * @param newValue
	 *            New value of the property
	 * @see PropertyChangeMessage
	 */
	protected void notify(String propertyName, Object oldValue, Object newValue) {
		this.notificationSupport.notify(propertyName, oldValue, newValue);
	}

	private void readObject(ObjectInputStream in) throws IOException,
			ClassNotFoundException {
		in.defaultReadObject();
		this.notificationSupport = new BeanNotificationSupport(this);
	}

	/**
	 * @see BeanNotificationSupport#setHasBufferedChanges(boolean)
	 */
	protected void setHasBufferedChanges(boolean hasBufferedChanges) {
		this.notificationSupport.setHasBufferedChanges(hasBufferedChanges);
	}

	/**
	 * @see BeanNotificationSupport#setNotificationMode(NotificationMode)
	 */
	public synchronized NotificationMode setNotificationMode(
			NotificationMode mode) {
		return this.notificationSupport.setNotificationMode(mode);
	}

	/**
	 * Sets the parent bean for this bean.
	 * 
	 * @param parentBean
	 *            New parent bean
	 */
	public void setParentBean(IParentBean parentBean) {
		this.parentBean = parentBean;
	}

	/**
	 * The notification is disabled.
	 * 
	 * @return Previous notification mode
	 */
	public NotificationMode stopNotification() {
		return setNotificationMode(NotificationMode.STOPPED);
	}

}
