package net.sf.doolin.bus.bean;

import java.util.Collections;
import java.util.LinkedList;

import net.sf.doolin.bus.Bus;
import net.sf.doolin.bus.support.PropertyChangeMessage;

import org.apache.commons.lang.Validate;

/**
 * This class defines an object that can notify the changes of an object to the
 * bus.
 * <p>
 * Notification of changes can be done on the spot, or buffered or completely
 * stopped.
 * <p>
 * TODO Any {@link BeanNotificationSupport} can have a parent and we can
 * therefore define a hierarchy of {@linkplain BeanNotificationSupport beans}.
 * Each level in the hierarchy participates in the notification of changes.
 * <p>
 * For example, in the following hierarchy:
 * 
 * <pre>
 * A {
 * 	B b;
 * }
 * 
 * B {
 * 	C c;
 * }
 * 
 * C {
 * 	String name;
 * }
 * </pre>
 * 
 * If C notifies a change on <code>name</code> as in:
 * 
 * <pre>
 * C {
 * 	...
 * 	notify(&quot;name&quot;, &quot;Old value&quot;, &quot;A new name&quot;);
 * 	...
 * }
 * </pre>
 * 
 * Following events are generated on the bus:
 * 
 * <pre>
 * C.name := &quot;A new name&quot;;
 * B.c.name := &quot;A new name&quot;;
 * A.b.c.name := &quot;A new name&quot;;
 * </pre>
 * 
 * @author Damien Coraboeuf
 * 
 */
public class BeanNotificationSupport {

	/**
	 * This property changes when changes are buffered.
	 */
	public static final String HAS_BUFFERED_CHANGES = "hasBufferedChanges";

	// private transient ParentBean parentBean;
	private final Object holder;

	private NotificationMode notificationMode = NotificationMode.NORMAL;

	private boolean cancel;
	private boolean cancelling;
	private boolean hasBufferedChanges;
	private LinkedList<PropertyChangeMessage> bufferedMessages = new LinkedList<PropertyChangeMessage>();

	public BeanNotificationSupport(Object holder) {
		this.holder = holder;
	}

	/**
	 * Sets the bean in BUFFERED notification mode.
	 * 
	 * @return Previous notification mode
	 */
	public NotificationMode bufferedNotification() {
		return setNotificationMode(NotificationMode.BUFFERED);
	}

	/**
	 * If the bean is in buffer mode, cancels all changes.
	 */
	public void cancelBuffer() {
		if (this.notificationMode == NotificationMode.BUFFERED
				&& this.bufferedMessages != null) {
			this.cancel = true;
			try {
				Collections.reverse(this.bufferedMessages);
				setNotificationMode(NotificationMode.NORMAL);
			} finally {
				this.cancel = false;
			}
		}
	}

	/**
	 * Cancels all buffered changes and restore the buffer state
	 */
	public void cancelBufferAndRestoreBuffer() {
		cancelBuffer();
		bufferedNotification();
	}

	// TODO Adds the Javadoc
	protected PropertyChangeMessage createMessage(String propertyName,
			Object oldValue, Object newValue) {
		PropertyChangeMessage pcm = new PropertyChangeMessage(this.holder,
				propertyName, oldValue, newValue);
		return pcm;
	}

	/**
	 * Sets the bean in NORMAL notification mode.
	 * 
	 * @return Previous notification mode
	 */
	public NotificationMode defaultNotification() {
		return setNotificationMode(NotificationMode.NORMAL);
	}

	/**
	 * Gets the current notification mode
	 * 
	 * @return Notification mode
	 */
	public NotificationMode getNotificationMode() {
		return this.notificationMode;
	}

	// TODO Javadoc
	public boolean isCancelling() {
		return this.cancelling;
	}

	/**
	 * @return the hasBufferedChanges
	 */
	public boolean isHasBufferedChanges() {
		return this.hasBufferedChanges;
	}

	/**
	 * Notifies to the bus that a property has changed. If the notification mode
	 * is BUFFERED, the change will be actually sent when notification mode is
	 * restored to NORMAL.
	 * 
	 * @param propertyName
	 *            Name of the property
	 * @param oldValue
	 *            Old value of the property
	 * @param newValue
	 *            New value of the property
	 * @see PropertyChangeMessage
	 */
	public void notify(String propertyName, Object oldValue, Object newValue) {
		if (!this.cancelling) {
			if (this.notificationMode == NotificationMode.BUFFERED) {
				PropertyChangeMessage pcm = createMessage(propertyName,
						oldValue, newValue);
				this.bufferedMessages.add(pcm);
				setHasBufferedChanges(true);
			} else if (this.notificationMode == NotificationMode.NORMAL) {
				PropertyChangeMessage pcm = createMessage(propertyName,
						oldValue, newValue);
				Bus.get().publish(pcm);
				// Notification to the parent
				if (this.holder instanceof IBean) {
					IBean bean = (IBean) this.holder;
					// Parent change for the children
					if (oldValue instanceof IBean) {
						((IBean) oldValue).setParentBean(null);
					}
					if (newValue instanceof IBean) {
						((IBean) newValue).setParentBean(new ParentBean(
								this.holder, propertyName));
					}
					// Parent notification
					IParentBean parentBean = bean.getParentBean();
					if (parentBean != null) {
						parentBean.notifyForChild(propertyName, oldValue,
								newValue);
					}
				}
			}
		}
	}

	/**
	 * @param hasBufferedChanges
	 *            the hasBufferedChanges to set
	 */
	protected void setHasBufferedChanges(boolean hasBufferedChanges) {
		boolean old = this.hasBufferedChanges;
		this.hasBufferedChanges = hasBufferedChanges;
		// Bypasses the Bean mechanism
		Bus.get().publish(
				new PropertyChangeMessage(this.holder, HAS_BUFFERED_CHANGES,
						old, hasBufferedChanges));
	}

	/**
	 * Sets the bean in the given notification mode. If the previous
	 * notification mode was BUFFERED, all buffered changes are now sent to the
	 * bus.
	 * 
	 * @param mode
	 *            Notification mode
	 * @return Previous notification mode.
	 */
	public synchronized NotificationMode setNotificationMode(
			NotificationMode mode) {
		// Validation of parameters
		Validate.notNull(mode, "Notification mode must not be null");
		// Old value
		NotificationMode old = this.notificationMode;
		// New mode
		this.notificationMode = mode;
		// Bufferisation ?
		if (old == NotificationMode.BUFFERED && mode == NotificationMode.NORMAL
				&& !this.bufferedMessages.isEmpty()) {
			for (PropertyChangeMessage message : this.bufferedMessages) {
				if (this.cancel) {
					this.cancelling = true;
					try {
						message.cancel();
					} finally {
						this.cancelling = false;
					}
				} else {
					notify(message.getPropertyName(), message
							.getOldPropertyValue(), message.getPropertyValue());
				}
			}
		}
		// Clears the bufferisation
		this.bufferedMessages.clear();
		setHasBufferedChanges(false);
		// OK
		return old;
	}

	/**
	 * The notification is disabled.
	 * 
	 * @return Previous notification mode
	 */
	public NotificationMode stopNotification() {
		return setNotificationMode(NotificationMode.STOPPED);
	}

}
