/**
 * 
 */
package net.sf.doolin.bus.cglib;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicReference;

import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;
import net.sf.doolin.bus.annotation.Property;
import net.sf.doolin.bus.bean.BeanNotificationSupport;
import net.sf.doolin.bus.bean.IBean;
import net.sf.doolin.bus.bean.IParentBean;
import net.sf.doolin.bus.bean.ParentBean;
import net.sf.doolin.util.Utils;

import org.apache.commons.lang.StringUtils;

public final class BeanMethodInterceptor<T> implements MethodInterceptor {
	public static String getNameFromMethod(String methodName, String prefix) {
		return StringUtils.uncapitalize(StringUtils.difference(prefix,
				methodName));
	}

	public static String getPropertyGetter(Method method, Object[] args) {
		// Looks for annotations
		Property propertyAnnotation = method.getAnnotation(Property.class);
		if (propertyAnnotation != null) {
			// The method must look like a getter
			String returnType = method.getReturnType().getName();
			if (!"void".equals(returnType) && args.length == 0) {
				// Property name
				String propertyName = getPropertyName(propertyAnnotation,
						method, "get", "is");
				// OK
				return propertyName;
			}
		}
		// Nothing has been found
		return null;
	}

	public static String getPropertyName(Property propertyAnnotation,
			Method method, String... prefixes) {
		String value = propertyAnnotation.value();
		if (StringUtils.isNotBlank(value)) {
			return value;
		} else {
			String methodName = method.getName();
			for (String prefix : prefixes) {
				if (methodName.startsWith(prefix)) {
					return getNameFromMethod(methodName, prefix);
				}
			}
			// Invalid method name
			throw new IllegalStateException(
					String
							.format(
									"Cannot find property name from %s method. You should provide a value for the @Property annotation.",
									methodName));
		}
	}

	public static String getPropertySetter(Method method, Object[] args) {
		// Looks for annotations
		Property propertyAnnotation = method.getAnnotation(Property.class);
		if (propertyAnnotation != null) {
			// The method must look like a setter
			String returnType = method.getReturnType().getName();
			if ("void".equals(returnType) && args.length == 1) {
				// Property name
				String propertyName = getPropertyName(propertyAnnotation,
						method, "set");
				// OK
				return propertyName;
			}
		}
		// Nothing has been found
		return null;
	}

	private final T instance;

	// Notification support
	private final AtomicReference<BeanNotificationSupport> support = new AtomicReference<BeanNotificationSupport>();

	// Hierarchy
	private final AtomicReference<IParentBean> parentBeanRef = new AtomicReference<IParentBean>();

	// Map of proxied properties
	private final Map<String, ProxyLink> proxiedProperties = new ConcurrentHashMap<String, ProxyLink>();

	public BeanMethodInterceptor(T instance) {
		this.instance = instance;
	}

	protected Object call(Object instance, Method method, Object[] args)
			throws Throwable {
		if (IBean.class.isAssignableFrom(method.getDeclaringClass())) {
			String methodName = method.getName();
			if ("getBeanNotificationSupport".equals(methodName)) {
				return this.support.get();
			} else if ("getParentBean".equals(methodName)) {
				return this.parentBeanRef.get();
			} else if ("setParentBean".equals(methodName)) {
				IParentBean parentBean = (IParentBean) args[0];
				this.parentBeanRef.set(parentBean);
				return null;
			} else {
				throw new UnsupportedOperationException(methodName);
			}
		} else {
			// Default behaviour
			return method.invoke(instance, args);
		}
	}

	public AtomicReference<BeanNotificationSupport> getSupport() {
		return this.support;
	}

	@Override
	public Object intercept(Object obj, Method method, Object[] args,
			MethodProxy proxy) throws Throwable {
		// Setter method
		String propertyName = getPropertySetter(method, args);
		if (StringUtils.isNotBlank(propertyName)) {
			// Call with notification
			return notify(propertyName, method, args);
		}
		// Getter method
		propertyName = getPropertyGetter(method, args);
		if (StringUtils.isNotBlank(propertyName)) {
			// Creates a proxy for the property value
			return proxy(obj, propertyName, method, args);
		}
		// Call without notification
		return call(this.instance, method, args);
	}

	protected Object notify(String propertyName, Method method, Object[] args)
			throws IllegalArgumentException, IllegalAccessException,
			InvocationTargetException {
		// Gets the old property value
		Object oldValue = Utils.getProperty(this.instance, propertyName);
		// Gets the new value
		Object newValue = args[0];
		// Default call
		Object returnedValue = method.invoke(this.instance, args);
		// Notification
		this.support.get().notify(propertyName, oldValue, newValue);
		// OK
		return returnedValue;
	}

	protected Object proxy(Object parentProxy, String propertyName,
			Method method, Object[] args) throws IllegalArgumentException,
			IllegalAccessException, InvocationTargetException {
		// Target
		Object target = method.invoke(this.instance, args);
		// Gets the already proxied item
		ProxyLink link = this.proxiedProperties.get(propertyName);
		if (link != null) {
			// Checks that targets are identical
			if (link.getTarget() == target) {
				return link.getProxy();
			}
		}
		// Not proxied yet or no longer valid, needs to create the proxy
		if (target == null) {
			// Clears the proxy
			this.proxiedProperties.remove(propertyName);
			// Just returns null
			return null;
		}
		// Proxy creation
		else {
			Object proxy = BeanProxyFactory.proxy(target);
			// Sets the parent
			if (proxy instanceof IBean) {
				((IBean) proxy).setParentBean(new ParentBean(parentProxy,
						propertyName));
			}
			// Registers the link
			this.proxiedProperties.put(propertyName, new ProxyLink(target,
					proxy));
			// OK
			return proxy;
		}
	}
}