package net.sf.doolin.bus.expression;

import java.util.Locale;

import net.sf.doolin.bus.support.PropertyChangeSupport;
import net.sf.doolin.bus.support.SubscriberValidator;
import net.sf.doolin.util.Utils;
import net.sf.doolin.util.expression.Expression;
import net.sf.doolin.util.expression.KeyToken;
import net.sf.doolin.util.expression.TextToken;
import net.sf.doolin.util.expression.TokenVisitor;
import net.sf.doolin.util.expression.ValueToken;

/**
 * A {@link BusExpression} is an {@link Expression} which is linked to a bean
 * model, its value may change as the model changes. The {@link #isVariable()}
 * method can be called to check if such changes can happen (for example,
 * expressions containing only leys and text cannot change).
 * 
 * <p>
 * To listens to the changes, one can do:
 * 
 * <pre>
 * BusExpression expression = ...
 * String initialValue = expression.getValue();
 * if (expression.isVariable()) {
 * 	expression.getValueModel().addValueChangeListener(new PropertyChangeListener () {
 * 		public void propertyChange(PropertyChangeEvent evt) {
 * 			String newExpressionValue = (String) evt.getNewValue();
 * 			...
 * 		} 
 * 	});
 * }
 * </pre>
 * 
 * @author Damien Coraboeuf
 */
public class BusExpression extends Expression {

	/**
	 * <code>value</code> property.
	 */
	public static final String VALUE = "value";

	/**
	 * Builds an expression with a fixed text.
	 * 
	 * @param locale
	 *            Locale for this expression
	 * @param text
	 *            Fixed value for this expression (not parsed)
	 */
	public BusExpression(Locale locale, String text) {
		super(locale, text);
	}

	/**
	 * Parses the expression
	 * 
	 * @param locale
	 *            Locale for the expression
	 * @param validator
	 *            Validator for the context
	 * @param context
	 *            Context object
	 * @param pattern
	 *            Pattern used to build the expression
	 */
	public BusExpression(Locale locale, final SubscriberValidator validator,
			final Object context, String pattern) {
		super(locale, context, pattern);
		// Checks if the expression has some variable parts
		if (isVariable() && validator != null) {
			// Tasks to register for each token
			final Runnable tokenTask = new Runnable() {

				@Override
				public void run() {
					evaluate();
				}
			};
			// Registers the expression for the bean
			visitTokensWith(new TokenVisitor() {

				@Override
				public void visitKeyToken(KeyToken keyToken) {
					// Nothing to register
				}

				@Override
				public void visitTextToken(TextToken textToken) {
					// Nothing to register
				}

				@Override
				public void visitValueToken(ValueToken valueToken) {
					if (valueToken.isBound()) {
						PropertyChangeSupport.subscribe(validator, context,
								valueToken.getPropertyPath(), tokenTask);
					}
				}
			});
		}
		// Evaluates the first value
		evaluate();
	}

	/**
	 * Builds an expression with a fixed text and for the default locale
	 * 
	 * @param text
	 *            Fixed text (not parsed)
	 * @see Locale#getDefault()
	 */
	public BusExpression(String text) {
		this(Locale.getDefault(), text);
	}

	/**
	 * Parses the expression
	 * 
	 * @param validator
	 *            Validator for the context
	 * @param context
	 *            Context object
	 * @param pattern
	 *            Pattern used to build the expression
	 */
	public BusExpression(SubscriberValidator validator, Object context,
			String pattern) {
		this(Locale.getDefault(), validator, context, pattern);
	}

	/**
	 * Connects the expression value to a bean
	 * 
	 * @param validator
	 *            Validator of the subscription
	 * @param bean
	 *            Target
	 * @param propertyPath
	 *            Target property
	 */
	public void connect(SubscriberValidator validator, Object bean,
			String propertyPath) {
		// Initial value
		Utils.setProperty(bean, propertyPath, getValue());
		// Subscription
		if (isVariable()) {
			PropertyChangeSupport.connectOneWayAndUpdate(validator, this,
					VALUE, bean, propertyPath);
		}
	}

	/**
	 * Internal method that sets a new value to the expression
	 * 
	 * @param newValue
	 *            New value
	 */
	@Override
	protected void setValue(String newValue) {
		String oldValue = getValue();
		super.setValue(newValue);
		PropertyChangeSupport.change(this, VALUE, oldValue, newValue);
	}

	/**
	 * Subscribes to the changes to the expression
	 * 
	 * @param validator
	 *            Validator of the subscriber
	 * @param runnable
	 *            Task to run in case of change
	 */
	public void subscribe(SubscriberValidator validator, Runnable runnable) {
		if (isVariable()) {
			PropertyChangeSupport.subscribe(validator, this, VALUE, runnable);
		}
	}
}
