package net.sf.doolin.bus.support;


/**
 * This subscriber accepts messages of a given class.
 * 
 * @author Damien Coraboeuf
 * 
 * @param <M>
 *            Type of message
 */
public abstract class AbstractClassSubscriber<M> extends AbstractSubscriber<M> {

	private final Class<M> messageClass;

	/**
	 * Builds a subscriber with the {@linkplain DefaultSubscriberValidator
	 * default validator}.
	 * 
	 * @param messageClass
	 *            Class for the messages
	 */
	public AbstractClassSubscriber(Class<M> messageClass) {
		this.messageClass = messageClass;
	}

	/**
	 * Builds a subscriber with a given validator.
	 * 
	 * @param messageClass
	 *            Class for the messages
	 * @param validator
	 *            Validator to use for this subscriber.
	 */
	public AbstractClassSubscriber(Class<M> messageClass,
			SubscriberValidator validator) {
		super(validator);
		this.messageClass = messageClass;
	}

	@Override
	@SuppressWarnings("unchecked")
	public final boolean accept(Object message) {
		return message != null
				&& this.messageClass.isAssignableFrom(message.getClass())
				&& acceptMessage((M) message);
	}

	/**
	 * This method is called to refine the acceptance after the message class
	 * has been validated. By default, it returns <code>true</code>.
	 * 
	 * @param message
	 *            Message to validate.
	 * @return Result for the acceptance
	 */
	protected boolean acceptMessage(M message) {
		return true;
	}

	@Override
	public String getTriggerDescription() {
		return String.format("class<%s>", this.messageClass.getSimpleName());
	}

}
