/**
 * 
 */
package net.sf.doolin.bus.support;

import java.util.concurrent.atomic.AtomicBoolean;

import net.sf.doolin.util.Utils;

/**
 * This subscriber connects two properties in two different beans, in both
 * directions. If the property of the source bean changes, then the property on
 * the target bean is updated and vice-versa.
 * 
 * @author Damien Coraboeuf
 * 
 */
public final class BothWayPropertyChangeSubscriber extends
		AbstractSubscriber<PropertyChangeMessage> {
	private final Object sourceBean;
	private final String sourcePropertyPath;
	private final Object targetBean;
	private final String targetPropertyPath;

	private final AtomicBoolean updating = new AtomicBoolean(false);

	/**
	 * Constructor.
	 * 
	 * @param validator
	 *            Validator for the subscription
	 * @param sourceBean
	 *            Source bean
	 * @param sourcePropertyPath
	 *            Source property path
	 * @param targetBean
	 *            Target bean
	 * @param targetPropertyPath
	 *            Target property path
	 */
	public BothWayPropertyChangeSubscriber(SubscriberValidator validator,
			Object sourceBean, String sourcePropertyPath, Object targetBean,
			String targetPropertyPath) {
		super(validator);
		this.sourceBean = sourceBean;
		this.sourcePropertyPath = sourcePropertyPath;
		this.targetBean = targetBean;
		this.targetPropertyPath = targetPropertyPath;
	}

	@Override
	public boolean accept(Object message) {
		if (message instanceof PropertyChangeMessage) {
			PropertyChangeMessage pm = (PropertyChangeMessage) message;
			return !this.updating.get()
					&& ((pm.getBean() == this.sourceBean && this.sourcePropertyPath
							.equals(pm.getPropertyName())) || (pm.getBean() == this.targetBean && this.targetPropertyPath
							.equals(pm.getPropertyName())));
		} else {
			return false;
		}
	}

	@Override
	public String getExecutionDescription() {
		return String.format("%s.%s <-> %s.%s", Utils.simpleToString(
				this.sourceBean, "<null>"), this.sourcePropertyPath, Utils
				.simpleToString(this.targetBean, "<null>"),
				this.targetPropertyPath);
	}

	@Override
	public String getTriggerDescription() {
		return getExecutionDescription();
	}

	@Override
	public void receive(PropertyChangeMessage message) {
		if (!this.updating.get()) {
			if (message.getBean() == this.sourceBean
					&& this.sourcePropertyPath
							.equals(message.getPropertyName())) {
				this.updating.set(true);
				try {
					Object sourcePropertyValue = message.getPropertyValue();
					Utils.setProperty(this.targetBean, this.targetPropertyPath,
							sourcePropertyValue);
				} finally {
					this.updating.set(false);
				}
			} else if (message.getBean() == this.targetBean
					&& this.targetPropertyPath
							.equals(message.getPropertyName())) {
				this.updating.set(true);
				try {
					Object targetPropertyValue = message.getPropertyValue();
					Utils.setProperty(this.sourceBean, this.sourcePropertyPath,
							targetPropertyValue);
				} finally {
					this.updating.set(false);
				}
			}
		}
	}

	@Override
	public String toString() {
		String sourceName = this.sourceBean.getClass().getSimpleName();
		String targetName = this.targetBean.getClass().getSimpleName();
		return "[" + sourceName + "." + this.sourcePropertyPath + "<=>"
				+ targetName + "." + this.targetPropertyPath + "]";
	}
}