package net.sf.doolin.bus.support;

import net.sf.doolin.util.Utils;

import org.apache.commons.lang.StringUtils;

/**
 * This subscriber listens to property changes on a particular bean.
 * 
 * @author Damien Coraboeuf
 * 
 */
public abstract class PropertyChangeSubscriber extends
		AbstractClassSubscriber<PropertyChangeMessage> {

	private final String executionDescription;
	private final Object bean;
	private final String propertyName;

	/**
	 * Constructor
	 * 
	 * @param executionDescription
	 *            Description of the action that is performed by this subscriber
	 * @param validator
	 *            Subscription validator
	 * @param bean
	 *            Bean to listen to
	 * @param propertyName
	 *            Property path to listen to or <code>null</code> if all
	 *            properties must be listened to
	 */
	public PropertyChangeSubscriber(String executionDescription,
			SubscriberValidator validator, Object bean, String propertyName) {
		super(PropertyChangeMessage.class, validator);
		this.executionDescription = executionDescription;
		this.bean = bean;
		this.propertyName = propertyName;
	}

	@Override
	public boolean acceptMessage(PropertyChangeMessage message) {
		return message.getBean() == this.bean
				&& (this.propertyName == null || this.propertyName
						.equals(message.getPropertyName()));
	}

	@Override
	public String getExecutionDescription() {
		return this.executionDescription;
	}

	@Override
	public String getTriggerDescription() {
		String p = this.propertyName != null ? this.propertyName : "any";
		String b = this.bean != null ? Utils.simpleToString(this.bean, "")
				: "any";
		String t = this.bean != null ? this.bean.getClass().getSimpleName()
				: "";
		return String.format("%s{%s}.%s", b, t, p);
	}

	/**
	 * Does the actual processing of the change
	 * 
	 * @param propertyName
	 *            Changed property
	 * @param propertyValue
	 *            New property value
	 */
	protected abstract void process(String propertyName, Object propertyValue);

	@Override
	public void receive(PropertyChangeMessage message) {
		process(message.getPropertyName(), message.getPropertyValue());
	}

	@Override
	public String toString() {
		String p = this.propertyName != null ? this.propertyName : "any";
		String b = this.bean != null ? String.format("%s{%s}", Utils
				.simpleToString(this.bean, ""), this.bean.getClass()
				.getSimpleName()) : "any";
		return String.format(
				"Listens for changes on %s property on %s bean and %s", p, b,
				StringUtils.uncapitalize(this.executionDescription));
	}

}
